import __future__
import warnings
import pandas as pd
import numpy as np
from tqdm import tqdm

warnings.filterwarnings("ignore", module="statsforecast.arima")
class Preprocessor():

    def __init__(
            self, 
            df: pd.DataFrame
        ) -> None:
        self.df = df.copy()
        self.df = self.df.sort_values("ds")
        self.df = self.df.reset_index(drop = True)
    
    def make_dataset(
            self,
            forecast_horizon: int, 
            season_length: int,
            date_freq: str = "MS",
            train_split: float = 0.8,
            models: list = [],
            fallback_model = None,
            verbose: bool = True
        ) -> tuple:

        self.forecast_horizon = forecast_horizon
        self.season_length = season_length
        self.date_freq = date_freq

        training_ids = np.random.choice(self.df.unique_id.unique(), int(len(self.df.unique_id.unique()) * train_split), replace = False)
        self.train_df = self.df[self.df.unique_id.isin(training_ids)]
        self.val_df = self.df[~self.df.unique_id.isin(training_ids)]
        dates = self.df['ds'].sort_values().unique()
        train_forecast_df = self.get_statistical_forecast(self.train_df[self.train_df['ds'].isin(dates[:-self.forecast_horizon])], models, fallback_model, verbose = verbose)
        val_forecast_df = self.get_statistical_forecast(self.val_df[self.val_df['ds'].isin(dates[:-self.forecast_horizon])], models, fallback_model, verbose = verbose)
        self.train_df = pd.merge(self.train_df, train_forecast_df, on = ["unique_id", "ds"], how = "left")
        self.val_df = pd.merge(self.val_df, val_forecast_df, on = ["unique_id", "ds"], how = "left")
        X_train, Y_train, ids_train, models = self.format_dataset(self.train_df)
        X_val, Y_val, ids_val, _ = self.format_dataset(self.val_df)

        return X_train, Y_train, ids_train, X_val, Y_val, ids_val, models

    def format_dataset(
            self, 
            local_df: pd.DataFrame
        ) -> tuple:
        pivot = pd.pivot_table(local_df, index = "unique_id", columns = "ds")
        models = np.unique(pivot.columns.get_level_values(0))
        x = []
        for model in models:
            if (model == "y"):
                x.append(np.expand_dims(pivot[model].iloc[:,:-self.forecast_horizon].to_numpy(), -1))
            else:
                x.append(np.expand_dims(pivot[model].to_numpy(), -1))
        y = np.expand_dims(pivot["y"].iloc[:,-self.forecast_horizon:].to_numpy(), -1)
        ids = pivot.index
        return x, y, ids, models

    def get_statistical_forecast(
            self, 
            local_df: pd.DataFrame, 
            models: list = [], 
            fallback_model = None,
            verbose: bool = True
        ) -> pd.DataFrame:
        groups = local_df.groupby("unique_id")
        return_df = []
        for unique_id, group in tqdm(groups, total = len(groups), disable = not verbose):
            output_dates = pd.date_range(group.ds.max(), periods = self.forecast_horizon + 1, freq = self.date_freq)[1:]
            output_ids = [unique_id] * len(output_dates)
            for model in models:
                try:
                    model = model(season_length = self.season_length)
                    model.fit(group['y'].to_numpy())
                except:
                    if fallback_model is not None:
                        model = fallback_model(season_length = self.season_length)
                        model.fit(group['y'].to_numpy())
                    else:
                        raise Exception("Model fit failed. Please provide a fallback model.")
                return_df.append(pd.DataFrame({
                    "unique_id": output_ids,
                    "ds": output_dates,
                    model.__repr__(): model.predict(self.forecast_horizon)['mean'],
                }))
        return_df = pd.concat(return_df)
        return_df = return_df.groupby(['unique_id', 'ds'], as_index = False).sum()
        return return_df


import __future__

import os
import json
import optuna
import numpy as np
import tensorflow as tf
import matplotlib.pyplot as plt
from datetime import datetime

class TrueSight:

    def __init__(
            self,
            models: list,
            input_shape: list, 
            forecast_horizon: int,
        ) -> None:
        self.model_folder = 'best_model'
        self.hparams_folder = './hparams'
        os.makedirs(self.hparams_folder, exist_ok=True)
        self.input_shape = input_shape
        self.forecast_horizon = forecast_horizon
        self.models = models
        self.set_hparams()
        self.model = self.get_model(self.hparams)

    def get_model(
            self, 
            hparams: dict = {}
        ) -> tf.keras.Model:
        x_inputs = []
        x_outputs = []
        x = []
        for i in range(len(self.models)):
            x_inputs.append(tf.keras.layers.Input((self.input_shape[i], 1), name = f"input_{i}"))
            x.append(x_inputs[i])
            x[i] = tf.keras.layers.LayerNormalization(epsilon=1e-8)(x[i])
            x[i] = tf.keras.layers.Dense(hparams['hidden_size'], activation='selu', name=f"dense_input_{i}")(x[i])
            x[i] = tf.keras.layers.Dropout(hparams['dropout_rate'], name=f"dropout_{i}")(x[i], training = True)
            x[i] = tf.keras.layers.Conv1D(hparams['num_filters'], kernel_size=hparams['kernel_size'], activation='selu', padding='same', name=f"conv1d_{i}_1")(x[i])
            x[i] = tf.keras.layers.MaxPooling1D(pool_size=2, padding='same', name=f"max_pool_{i}_1")(x[i])
            x[i] = tf.keras.layers.Conv1D(hparams['num_filters'], kernel_size=hparams['kernel_size'], activation='selu', padding='same', name=f"conv1d_{i}_2")(x[i])
            x[i] = tf.keras.layers.MaxPooling1D(pool_size=2, padding='same', name=f"max_pool_{i}_2")(x[i])
            x[i] = tf.keras.layers.Bidirectional(tf.keras.layers.LSTM(hparams['lstm_units'], return_sequences=True), name=f"bidirectional_lstm_{i}")(x[i])
            x[i] = tf.keras.layers.TimeDistributed(tf.keras.layers.Dense(hparams['hidden_size'], activation='selu'), name=f"timedistributed_dense_{i}")(x[i])
            x[i] = tf.keras.layers.MultiHeadAttention(hparams['num_heads'], hparams['key_dim'], name = f"multihead_self_attention_{i}")(x[i], x[i])
            x[i] = tf.keras.layers.Flatten(name=f"flatten_{i}")(x[i])
            x_outputs.append(x[i])
        y = tf.keras.layers.Concatenate(name="concatenate")(x_outputs)
        y = tf.keras.layers.Dense(hparams['hidden_size'], activation='selu', name="dense_output")(y)
        y = tf.keras.layers.Dropout(hparams['dropout_rate'], name="dropout_output")(y, training = True)
        y = tf.keras.layers.Dense(self.forecast_horizon, activation='selu', name="output_dense")(y)
        model = tf.keras.Model(x_inputs, y, name="TrueSight")
        optimizer = tf.keras.optimizers.Adam(learning_rate=hparams['learning_rate'])
        model.compile(optimizer=optimizer, loss='mse')
        return model
    
    def fit(
            self, 
            X_train: list, 
            Y_train: np.ndarray, 
            X_val: list, 
            Y_val: np.ndarray, 
            batch_size: int = 128, 
            epochs: int = 100, 
            callbacks: list = [],
            save_best_model: bool = True,
            verbose: bool = True
        ):
        self.history = self.model.fit(
            x = X_train,
            y = Y_train,
            batch_size = batch_size,
            epochs = epochs,
            validation_data = [X_val, Y_val],
            callbacks = callbacks,
            verbose = verbose
            )
        if save_best_model: self.model.save(self.model_folder)

    def predict(
            self,
            X: list,
            n_repeats: int = 1,
            batch_size: int = 128,
            n_quantiles: int = 10,
            return_quantiles: bool = False,
            verbose: bool = True
        ) -> np.ndarray:

        yhat = []
        for i in range(n_repeats):
            yhat.append(self.model.predict(X, batch_size=batch_size, verbose=verbose))
        yhat = np.array(yhat)

        if return_quantiles: yhat = np.quantile(yhat, np.linspace(0, 1, n_quantiles), axis=0)
        else: yhat = np.mean(yhat, axis=0)
        return yhat

    def load_model(self):
        if not os.path.exists(self.model_folder): raise Exception("No model found")
        self.model = tf.keras.models.load_model(self.model_folder)
    
    def load_hparams(self):
        if not os.path.exists(f'{self.hparams_folder}/best_hparams.json'): raise Exception("No best hparams found, please run the auto_tune() first")
        with open(f'{self.hparams_folder}/best_hparams.json', 'r') as file: self.hparams = json.load(file)

    def set_hparams(
            self,
            num_filters: int = 64,
            kernel_size: int = 3,
            lstm_units: int = 64,
            hidden_size: int = 64,
            num_heads: int = 4,
            key_dim: int = 64,
            learning_rate: float = 0.001,
            dropout_rate: float = 0.2,
        ) -> None:
        self.hparams = {
            'num_filters': num_filters,
            'kernel_size': kernel_size,
            'lstm_units': lstm_units,
            'hidden_size': hidden_size,
            'num_heads': num_heads,
            'key_dim': key_dim,
            'learning_rate': learning_rate,
            'dropout_rate': dropout_rate,
        }

    def objective(self, trial):
        num_filters = trial.suggest_int("num_filters", self.min_num_filters, self.max_num_filters)
        kernel_size = trial.suggest_int("kernel_size", self.min_kernel_size, self.max_kernel_size)
        lstm_units = trial.suggest_int("lstm_units", self.min_lstm_units, self.max_lstm_units)
        hidden_size = trial.suggest_int("hidden_size", self.min_hidden_size, self.max_hidden_size)
        num_heads = trial.suggest_int("num_heads", self.min_num_heads, self.max_num_heads)
        key_dim = trial.suggest_int("key_dim", self.min_key_dim, self.max_key_dim)
        dropout_rate = trial.suggest_float("dropout_rate", self.min_dropout_rate, self.max_dropout_rate )
        learning_rate = trial.suggest_float("learning_rate", self.min_learning_rate, self.max_learning_rate)
        hparams = {
            'num_filters': num_filters,
            'kernel_size': kernel_size,
            'lstm_units': lstm_units,
            'hidden_size': hidden_size,
            'num_heads': num_heads,
            'key_dim': key_dim,
            'learning_rate': learning_rate,
            'dropout_rate': dropout_rate,
        }

        model = self.get_model(hparams)
        model.fit(self.X_train, self.Y_train, epochs = self.epochs, batch_size = self.batch_size, verbose = 0)
        score = model.evaluate(self.X_val, self.Y_val, batch_size = self.batch_size, verbose = 0)
        score = np.array(score).mean()
        with open(f'{self.hparams_folder}/{datetime.now().strftime("%Y%m%d%H%M%S")}-{score}.json', 'w') as file: json.dump(hparams, file)
        return score

    def auto_tune(
            self,
            X_train: list,
            Y_train: np.ndarray,
            X_val: list,
            Y_val: np.ndarray,
            n_trials: int,
            batch_size: int = 128,
            epochs: int = 5,
            min_num_filters: int = 32,
            max_num_filters: int = 256,
            min_kernel_size: int = 3,
            max_kernel_size: int = 21,
            min_lstm_units: int = 32,
            max_lstm_units: int = 256,
            min_hidden_size: int = 128,
            max_hidden_size: int = 1024,
            min_num_heads: int = 2,
            max_num_heads: int = 10,
            min_key_dim: int = 16,
            max_key_dim: int = 128,
            min_learning_rate: float = 0.0001,
            max_learning_rate: float = 0.01,
            min_dropout_rate: float = 0.05,
            max_dropout_rate: float = 0.4,
        ) -> dict:

        self.X_train = X_train
        self.Y_train = Y_train
        self.X_val = X_val
        self.Y_val = Y_val
        self.batch_size = batch_size
        self.epochs = epochs
        self.min_num_filters = min_num_filters
        self.max_num_filters = max_num_filters
        self.min_kernel_size = min_kernel_size
        self.max_kernel_size = max_kernel_size
        self.min_lstm_units = min_lstm_units
        self.max_lstm_units = max_lstm_units
        self.min_hidden_size = min_hidden_size
        self.max_hidden_size = max_hidden_size
        self.min_num_heads = min_num_heads
        self.max_num_heads = max_num_heads
        self.min_key_dim = min_key_dim
        self.max_key_dim = max_key_dim
        self.min_learning_rate = min_learning_rate
        self.max_learning_rate = max_learning_rate
        self.min_dropout_rate = min_dropout_rate
        self.max_dropout_rate = max_dropout_rate
        
        file_list = os.listdir(self.hparams_folder)
        for file_name in file_list:
            file_path = os.path.join(self.hparams_folder, file_name)
            if os.path.isfile(file_path):
                os.remove(file_path)

        study = optuna.create_study(direction="minimize")
        study.optimize(self.objective, n_trials=n_trials)
        self.hparams = study.best_trial.params
        with open(f'{self.hparams_folder}/best_hparams.json', 'w') as file: json.dump(self.hparams, file)
        return self.hparams

    def plot_history(self):
        if not hasattr(self, 'history'): raise Exception('No history found. Please train the model first.')
        plt.plot(self.history.history['loss'], label='train')
        plt.plot(self.history.history['val_loss'], label='validation')
        plt.legend()
        plt.show()


import __future__

import numpy as np
import matplotlib.pyplot as plt
from truesight.metrics import mse
from scipy.optimize import curve_fit

class AdditiveDecomposition():
    def __init__(
            self, 
            season_length: int, 
            smooth_factor: float = 0.6
        ) -> None:
        self.season_length = season_length
        self.smooth_factor = smooth_factor
        self.alias = "Additive Decomposition"
    
    def __repr__(
            self
        ) -> str:
        return self.alias

    def fit(
            self, 
            y: np.ndarray, 
            trend_estimators: list = ["linear", "log"]
        ) -> None:

        self.trend_estimators = []
        for trend_estimator in trend_estimators:
            match trend_estimator:
                case "linear":
                    self.trend_estimators.append(self.linear_eq)
                case "log":
                    self.trend_estimators.append(self.log_eq)
                case "power":
                    self.trend_estimators.append(self.power_eq)
                case "exp":
                    self.trend_estimators.append(self.exp_eq)
                case _:
                    raise ValueError("Trend estimator not found")

        self.y = y
        self.input_size = len(y)
        self.total_size = self.input_size * 2
        self.trend_line = self.get_trend()
        self.season_line = self.get_seasonality()
    
    def predict(
            self, 
            fcst_horizon: int
        ) -> dict:
        if (fcst_horizon < 1): raise ValueError("The forecast horizon must be greater than 0")
        if (fcst_horizon > self.input_size): raise ValueError("The forecast horizon cannot be greater than the input size")
        prediction = self.trend_line + self.season_line
        prediction[prediction < 0] = 0
        prediction = prediction[-fcst_horizon:]
        prediction_return = {}
        prediction_return['mean'] = prediction
        return prediction_return

    def get_trend(
            self
        ) -> np.ndarray:
        x = np.arange(1, self.total_size + 1)
        temp_trends = []
        for trend_estimator in self.trend_estimators:
            try:
                popt, _ = curve_fit(trend_estimator, x[:self.input_size], self.y)
                temp_trends.append(trend_estimator(x, *popt))
            except:
                pass
        best_score = np.inf
        best_trend = None
        for trend in temp_trends:
            score = mse(self.y, trend[:self.input_size], return_mean=True)
            if (score < best_score):
                best_trend = trend
                best_score = score
        trend_line = best_trend
        trend_line[trend_line < 0] = 0
        return trend_line

    def get_seasonality(
            self
        ) -> np.ndarray:
        seasonality = []
        dataset_season = self.y - self.trend_line[:self.input_size]
        cycles = int(self.input_size/self.season_length)
        for i in range(self.season_length):
            step_val = []
            for step in range(cycles):
                step_val.append(dataset_season[i + (step * self.season_length)])
            step_val = np.array(step_val).sum(axis = 0)
            step_val = np.nan_to_num(step_val / cycles)
            seasonality.append(step_val)
        seasonality = np.asarray(seasonality)
        seasonality = np.tile(seasonality, self.total_size)[:self.total_size]
        if (self.smooth_factor > 0): seasonality = self.smooth(seasonality, self.smooth_factor)
        return seasonality

    def plot_test(
            self, 
            fcst_horizon: int
        ) -> None:
        pred = self.prediction(fcst_horizon)
        _, ax = plt.subplots(figsize=(10, 6), dpi=100)
        range_total = np.arange(0, self.total_size)
        range_input = np.arange(0, self.input_size)
        range_output = np.arange(self.input_size, self.total_size)
        ax.plot(range_total, self.trend_line, label = "trend")
        ax.plot(range_input, self.y, label = "history")
        ax.plot(range_output, pred, label = "prediction")
        ax.plot(range_total, self.trend_line + self.season_line, label = "prediction")
        ax.plot(self.season_line, label = "seasonality")
        ax.legend()
        plt.show()

    def smooth(
            self, 
            scalars: np.ndarray, 
            weight: float
        ) -> list: 
        last = scalars[0]
        smoothed = list()
        for point in scalars:
            smoothed_val = last * weight + (1 - weight) * point 
            smoothed.append(smoothed_val)
            last = smoothed_val
        return smoothed

    def linear_eq(
            self, 
            x: np.ndarray, 
            a: float, 
            c: float
        ) -> np.ndarray:
        return a*x + c

    def log_eq(
            self, 
            x: np.ndarray, 
            a: float, 
            c: float
        ) -> np.ndarray:
        return a*np.log(x) + c

    def power_eq(
            self, 
            x: np.ndarray, 
            a: float,
            b: float,
            c: float
        ) -> np.ndarray:
        return a*np.power(x, b) + c
    
    def exp_eq(
            self, 
            x: np.ndarray, 
            a: float,
            b: float,
            c: float
        ) -> np.ndarray:
        return np.power(a, b * x) + c


import numpy as np
import pandas as pd

def get_input_shapes(X):
    input_shapes = []
    for x in X:
        input_shapes.append(x.shape[1])
    return input_shapes

def generate_syntetic_data(
        num_time_steps: int,
        seasonal_lenght: int,
        num_series: int,
    ) -> pd.DataFrame:
    dataset = []
    for _ in range(num_series):
        linear_slope = np.random.normal(0, 1)
        seasonal_amplitude = np.random.randint(1, 10)
        noise_amplitude = np.random.randint(1, 10)
        irregularity_amplitude = np.random.randint(1, 10)
        scale_factor = np.random.randint(1, 10)
        time = np.arange(num_time_steps)
        linear_trend = linear_slope * time
        seasonal_cycle = seasonal_amplitude * np.sin(2 * np.pi * time / seasonal_lenght)
        noise = np.random.normal(0, noise_amplitude, size=num_time_steps)
        irregularities = np.random.normal(0, irregularity_amplitude, size=num_time_steps)
        dataset.append(scale_factor * (linear_trend + seasonal_cycle + noise + irregularities))
    dataset = np.array(dataset)
    dataset[dataset < 0] = 0
    dates = pd.date_range(start='2020-01-01', periods=num_time_steps, freq='MS')
    ids = np.arange(dataset.shape[0])
    df = []
    for i, ts in enumerate(dataset):
        df.append(pd.DataFrame({"unique_id": np.tile(ids[i], len(ts)), "ds": dates, "y": ts}))
    return pd.concat(df)


from distutils.core import setup

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = 'truesight',
    packages = ['truesight'],
    version = 'v0.0.2a0',
    description = 'Truesight is a python package for time series prediction using deep learning and statistical models.',
    author = 'Renan Otvin Klehm',
    author_email = 'renanotivin@hotmail.com',
    url = 'https://github.com/renanklehm/true-sight',
    download_url = 'https://github.com/renanklehm/true-sight/archive/refs/tags/v0.0.1-alpha.tar.gz',
    keywords = ['time series', 'prediction'],
    classifiers = [],
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'tensorflow>=2.11.0',
        'optuna>=2.10.1',
        'numpy>=1.23.5',
        'pandas>=1.5.3',
        'matplotlib>=3.6.3',
        'scipy>=1.10.0',
        'tqdm>=4.64.1',
    ]
)


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

np.seterr(divide='ignore')

def bias(
        y_true: np.ndarray, 
        y_pred: np.ndarray,  
        axis: int = -1, 
        return_mean: bool = False
    ) -> np.ndarray:
    bias = y_true - y_pred
    if (return_mean):
        return np.mean(bias)
    else:
        return np.mean(bias, axis = axis)

def mae(
        y_true: np.ndarray, 
        y_pred: np.ndarray,  
        axis: int = -1, 
        return_mean: bool = False
    ) -> np.ndarray:
    mae = abs(y_true - y_pred)
    if (return_mean):
        return np.mean(mae)
    else:
        return np.mean(mae, axis = axis)  

def smape(
        y_true: np.ndarray, 
        y_pred: np.ndarray,  
        axis: int = -1, 
        return_mean: bool = False
    ) -> np.ndarray:
    smape = abs(y_true - y_pred)
    smape = smape / (abs(y_true) + abs(y_pred))
    smape = np.where((y_true==0) & (y_pred==0), np.zeros(smape.shape), smape)
    smape = np.where(np.isnan(smape), np.ones(smape.shape), smape)
    smape = np.where(np.isinf(smape), np.ones(smape.shape), smape)
    if (return_mean):
        return np.mean(smape)
    else:
        return np.mean(smape, axis = axis)

def mape(
        y_true: np.ndarray, 
        y_pred: np.ndarray,  
        axis: int = -1, 
        return_mean: bool = False
    ) -> np.ndarray:
    mape = abs(y_true - y_pred)
    mape = mape / y_true
    mape = np.where((y_true==0) & (y_pred==0), np.zeros(mape.shape), mape)
    mape = np.where(np.isnan(mape), np.ones(mape.shape), mape)
    mape = np.where(np.isinf(mape), np.ones(mape.shape), mape)
    if (return_mean):
        return np.mean(mape)
    else:
        return np.mean(mape, axis = axis)

def mse(
        y_true: np.ndarray, 
        y_pred: np.ndarray,  
        axis: int = -1, 
        return_mean: bool = False
    ) -> np.ndarray:
    mse = (y_true - y_pred) ** 2
    if (return_mean):
        return np.mean(mse)
    else:
        return np.mean(mse, axis = axis)

def rmse(
        y_true: np.ndarray, 
        y_pred: np.ndarray,  
        axis: int = -1, 
        return_mean: bool = False
    ) -> np.ndarray:
    rmse = (y_true - y_pred) ** 2
    if (return_mean):
        return np.sqrt(np.mean(rmse))
    else:
        return np.sqrt(np.mean(rmse, axis = axis))

class Evaluator:
    def __init__(
            self,
            x: np.ndarray,
            y_true: np.ndarray,
            y_pred: np.ndarray,
            ids: np.ndarray):
        self.x = x
        self.y_true = y_true
        self.y_pred = y_pred
        self.ids = ids

    def evaluate_prediction(
            self,
            evaluators: list = [],
            return_mean: bool = False
        ):
        df_list = []
        for i, id in enumerate(self.ids):
            df = {}
            for evaluator in evaluators:
                df[evaluator.__name__] = evaluator(self.y_true[i], self.y_pred.mean(axis=0)[i], return_mean = True)
            df_list.append(pd.DataFrame(df, index = [id]))
        if (return_mean):
            return pd.concat(df_list).mean()
        else:
            return pd.concat(df_list)

    def plot_exemple(self):
        idx = np.random.randint(0, len(self.y_true))

        ytrue = np.squeeze(self.y_true[idx])
        yhat = np.squeeze(self.y_pred.mean(axis=0)[idx])

        range_input = np.arange(0, self.x[-1].shape[1])
        range_output = np.arange(self.x[-1].shape[1], len(ytrue) + self.x[-1].shape[1])

        _, ax = plt.subplots(figsize=(10, 6), dpi=100)
        ax.plot(range_output, ytrue, ".-", color = "darkslategray", label = "True data")
        ax.plot(range_input, np.squeeze(self.x[-1][idx]), label = "Input", color = "darkslategray", alpha = 0.5)

        for i in range(yhat.shape[0] // 2):
            if (i == 0):
                ax.fill_between(range_output, np.squeeze(self.y_pred[-i-1,idx]), np.squeeze(self.y_pred[i,idx]), alpha = 0.3 / (self.y_pred.shape[0] / 2), color = "brown")
            else:
                ax.fill_between(range_output, np.squeeze(self.y_pred[-i-1,idx]), np.squeeze(self.y_pred[i,idx]), alpha = 0.3 / (self.y_pred.shape[0] / 2), color = "brown")
        ax.plot(range_output, yhat, "v-", label = "TrueSight Prediction", color = "brown")
        ax.axvline(x = range_output[0], linestyle = "--", color = 'crimson', label = 'Training Data')
        ax.set_ylim(0)

        plt.legend()
        plt.tight_layout()
        plt.show()